from functools import partial

from tqdm import tqdm

from simplepipeline.map.base_map import Map, MultiMap


def seq_map_(func, iterable, **kwargs):
    """
        Args:
        func: Function to be executed
        iterable: Iterable to be looped over
        **kwargs: Arguments of the function
    Returns:
        Result of the function
    """
    return [func().run_map(x, **kwargs) for x in tqdm(iterable)]


def seq_map(func, iterable, **kwargs):
    return Map(partial(seq_map_, func, iterable, **kwargs), func)


def seq_map_unpack(func, iterable, **kwargs):
    """
    Used for funcs that return several values (or a tuple of values). Note that any return
    value is a list.
    """
    return MultiMap(seq_map(func, iterable, **kwargs))



import logging

logger = logging.getLogger("root")


class Pipeline:

    def __init__(self, name=None):
        self.name = name
        self._tasks = []

    def register(self, task_name):
        """
        Registers a task. If a task is executed several times in a row, only one entry is created.
        """
        if not self._tasks or self._tasks[len(self._tasks) - 1] != task_name:
            self._tasks.append(task_name)

    def executed_tasks(self):
        logger.info(f"Pipeline: {'Not Specified' if self.name is None else self.name}")
        _ = [logger.info(f"\t Task: {task}") for task in self._tasks]
        return self._tasks



import logging
from datetime import timedelta
from timeit import default_timer as timer

logger = logging.getLogger("root")


class TaskLogger:

    def __init__(self, func_name):
        self.func_name = func_name
        self._start = None
        self._end = None

    def start(self):
        logger.info(f"Starting: {self.func_name}")
        self._start = timer()

    def stop(self, additional=""):
        self._end = timer()
        logger.info(f"Finished: {self.func_name}. "
                    f"Time elapsed: {str(timedelta(seconds=self._end - self._start)).split('.')[0]}. "
                    f"{additional}")



from functools import partial

import ray

from simplepipeline.map.base_map import Map, MultiMap


@ray.remote
def run(obj, *args, **kwargs):
    return obj.run_map(*args, **kwargs)


def par_map_(func, iterable, **kwargs):
    """
    Lot of overhead because of multiprocessing. Should only be used for processing intensive tasks.
    Args:
        func: Function to be executed
        iterable: Iterable to be looped over
        **kwargs: Arguments of the function
    Returns:
        Result of the function
    """
    ray.init()
    futures = [run.remote(func(), i, **kwargs) for i in iterable]
    results = ray.get(futures)
    ray.shutdown()
    return results


def par_map(func, iterable, **kwargs):
    return Map(partial(par_map_, func, iterable, **kwargs), func)


def par_map_unpack(func, iterable, **kwargs):
    """
    Used for funcs that return several values (or a tuple of values). Note that any return
    value is a list.
    """
    return MultiMap(par_map(func, iterable, **kwargs))



from simplepipeline.utils.task_logger import TaskLogger


class Map:

    def __init__(self, partial_func, task):
        self.partial_func = partial_func
        self.task = task()

    def run(self):
        task_logger = TaskLogger(self.task.func_name)
        task_logger.start()
        result = self.partial_func()
        task_logger.stop()
        return result


class MultiMap:

    def __init__(self, func):
        self.func = func

    def run(self):
        return map(list, zip(*self.func.run()))



from distutils.core import setup
from setuptools import find_packages

setup(
    name='simplepipeline',  # How you named your package folder (MyLib)
    version='v0.2.1',  # Start with a small number and increase it with every change you make
    packages=find_packages(),
    include_package_data=True,
    license='MIT',  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description='',  # Give a short description about your library
    author='Lucas Fobian',  # Type in your name
    # author_email = 'your.email@domain.com',      # Type in your E-Mail
    # Provide either the link to your github or to your website
    url='https://github.com/lucasfbn/SimplePipeline',
    # I explain this later on
    keywords=[],  # Keywords that define your package best
    install_requires=[  # I get to this in a second
        'pandas',
        'ray',
        'tqdm'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',  # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',  # Again, pick a license
        # Specify which pyhton versions that you want to support
        'Programming Language :: Python :: 3',
    ],
)



from functools import partial

from simplepipeline.pipeline import Pipeline
from simplepipeline.utils.task_logger import TaskLogger


class Task:
    PIPELINE = None

    def __init__(self, func, func_name, args, kwargs):
        self.func = func
        self.func_name = func_name
        self.func_partial = partial(func, *args, **kwargs)

        self.register()

    def run_map(self, *args, **kwargs):
        result = self.func(*args, **kwargs)
        return result

    def run(self):
        task_logger = TaskLogger(self.func_name)
        task_logger.start()
        result = self.func_partial()
        task_logger.stop()
        return result

    def register(self):
        if self.PIPELINE is not None:
            self.PIPELINE.register(self.func_name)


class FilterTask(Task):

    def __init__(self, func, func_name, args, kwargs):
        super().__init__(func, func_name, args, kwargs)
        self.in_len = len(args[0])

    def run(self):
        task_logger = TaskLogger(self.func_name)
        task_logger.start()
        result = self.func_partial()

        if isinstance(result, tuple):
            out_len = len(result[0])
        else:
            out_len = len(result)

        task_logger.stop(f"Dropped {self.in_len - out_len} items.")
        return result


def task(func):
    def wrapper(*args, **kwargs):
        return Task(func, func.__name__, args, kwargs)

    return wrapper


def filter_task(func):
    def wrapper(*args, **kwargs):
        return FilterTask(func, func.__name__, args, kwargs)

    return wrapper


def set_pipeline(pipeline: Pipeline):
    Task.PIPELINE = pipeline


def get_pipeline():
    return Task.PIPELINE


def delete_pipeline():
    Task.PIPELINE = None
